from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import warnings


__all__ = [
    'Float',
    'RGBColour',
]


class Float(object):
    def __init__(self):
        warnings.warn("This class will be removed.",
                      DeprecationWarning)


class RGBColour(object):
    def __init__(self):
        warnings.warn("This class will be removed.",
                      DeprecationWarning)
