"""
********************************************************************************
geometry
********************************************************************************

.. currentmodule:: compas_blender.geometry

Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:


Plugins
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    booleans.boolean_difference_mesh_mesh
    booleans.boolean_intersection_mesh_mesh
    booleans.boolean_union_mesh_mesh

"""
